<?php declare(strict_types=1);

namespace Infomaker\Imengine;

use Infomaker\Everyware\Support\Format;
use Infomaker\Imengine\CropData\Interfaces\AutoCropDataInterface;
use Infomaker\Imengine\CropData\Interfaces\CropResizeDataInterface;
use Infomaker\Imengine\CropData\Interfaces\NewsPilotCropDataInterface;
use Infomaker\Imengine\Helpers\ImengineHelper;
use RuntimeException;

/**
 * Class Imengine
 *
 * @package Infomaker\Imengine
 */
class Imengine
{
    /**
     * Contains the base url to the Imengine server
     *
     * @var string
     */
    private static $serverUrl;

    /**
     * Contains an array of the parameters that will be sent to Imengine server
     *
     * @var array
     */
    private $params;

    /**
     * Contains the default parameters that's always sent to Imengine regardless of the function used
     *
     * @var array
     */
    private static $defaultParams = [
        'function' => null,
        'type'     => 'preview', // preview|primary
        'source'   => 'false',
        'q'        => 75,
    ];

    /**
     * Imengine constructor.
     *
     * @param string $serverUrl
     * @param array  $params
     */
    private function __construct(string $serverUrl, array $params)
    {
        self::$serverUrl = $serverUrl;
        $this->params    = array_replace(static::$defaultParams, $params);
    }

    /**
     * Setup Imengine by serve it with server url
     *
     * @param string $serverUrl
     */
    public static function setup(string $serverUrl): void
    {
        static::$serverUrl = $serverUrl;
    }

    /**
     * Use an uuid to specify witch image to manipulate.
     * (The image must exist in the Open content configured in the Imengine api in use)
     *
     * @param string $uuid
     *
     * @return string
     */
    public function fromUuid(string $uuid): string
    {
        return $this->buildFromParams(array_merge(['uuid' => $uuid], $this->params));
    }

    /**
     * Use an external image to manipulate.
     * (The new version will be saved on Imengine server)
     *
     * @param string $imgSrc
     *
     * @return string
     */
    public function fromSrc(string $imgSrc): string
    {
        // Add format if not already set
        if ( ! isset($this->params['format'])) {
            $this->params['format'] = pathinfo(strtok($imgSrc, '?'), PATHINFO_EXTENSION);
        }

        return $this->buildFromParams(array_merge(['imgsrc' => trim($imgSrc)], $this->params));
    }

    // Chainable functions: functions to chain on mirror functions to further manipulate your query before sending it to Imengine
    // ======================================================

    /**
     * Add or override parameters not specific to the function being used
     *
     * @param array $params
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public function with(array $params): Imengine
    {
        $this->params = array_replace($this->params, $params);

        return $this;
    }

    /**
     * Function to allow for quality change of the requested image
     *
     * @param int $quality [1-100]
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public function quality(int $quality): Imengine
    {
        if ($quality > 0 && $quality <= 100) {
            $this->params = array_replace($this->params, ['q' => $quality]);
        }

        return $this;
    }

    /**
     * Function to set format of the requested image
     *
     * @param string $format [jpg,png]
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public function format(string $format): Imengine
    {
        if (in_array($format, ['jpg', 'png'])) {
            $this->params = array_replace($this->params, ['format' => $format]);
        }

        return $this;
    }

    // Protected functions:
    // ======================================================

    /**
     * Function to build the query url to Imengine API
     *
     * @param array $params
     *
     * @return string
     */
    protected function buildFromParams(array $params): string
    {
        return static::$serverUrl . '?' . http_build_query($params);
    }

    /**
     * Function to create an instance with only params.
     * Used by all Imengine server-function mirrored functions
     *
     * @param string $func
     * @param array  $params
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function createMirrorFunction(string $func, array $params = null): Imengine
    {
        return new static(static::getBaseUrl(), array_merge(['function' => $func], $params ?? []));
    }

    /**
     * Function to get DEV or PROD img-base URL (with or without CDN)
     *
     * @return string
     */
    protected static function getImengineUrl(): string
    {
        if ( ! static::$serverUrl) {
            throw new RuntimeException('You need to setup Imengine through Imengine::setup to serve images');
        }

        return static::$serverUrl;
    }

    // Mirror functions: mirrors the functions found on Imengine API
    // ======================================================

    /**
     * Scale image to cover the limits.
     *
     * @param int|null $width
     * @param int|null $height
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function cover(int $width = null, int $height = null): Imengine
    {
        return static::createMirrorFunction('cover', compact('width', 'height'));
    }


    /**
     * Function to create a url from scratch
     *
     * @param array $params
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function create(array $params = null): Imengine
    {
        return new static(static::getBaseUrl(), $params ?? []);
    }

    /**
     * Get base url
     *
     * @return string
     */
    private static function getBaseUrl(): string
    {
        try {
            return static::getImengineUrl();
        } catch (RuntimeException $e) {
            /** @noinspection ForgottenDebugOutputInspection */
            error_log($e->getMessage());

            return '/';
        }
    }

    /**
     * Crop image into defined crop-size from custom start-positions.
     * Then scale into defined size
     *
     * @param \Infomaker\Imengine\CropData\Interfaces\CropResizeDataInterface $cropData
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function cropResize(CropResizeDataInterface $cropData): Imengine
    {
        return static::createMirrorFunction('cropresize', [
            'crop_w' => $cropData->getCropWidth(),
            'crop_h' => $cropData->getCropHeight(),
            'width'  => $cropData->getWidth(),
            'height' => $cropData->getHeight(),
            'x'      => $cropData->getCropX(),
            'y'      => $cropData->getCropY(),
        ]);
    }

    /**
     * Scale image to fit inside the limits.
     * Add scale up to make the image cover the maxsize-area
     * (Can exceed limits on scale up)
     *
     * @param int       $maxSize
     * @param bool|null $scaleUp
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function fit(int $maxSize, bool $scaleUp = null): Imengine
    {
        return static::createMirrorFunction('fit', [
            'maxsize' => $maxSize,
            'scaleup' => $scaleUp ?? true,
        ]);
    }

    /**
     * Crop image into defined size from center position
     *
     * @param $width
     * @param $height
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function hardCrop($width, $height): Imengine
    {
        return static::createMirrorFunction('hardcrop', compact('width', 'height'));
    }

    /**
     * Crop image into defined size from custom start-positions.
     * Implemented as crop for NewsPilot
     *
     * @param \Infomaker\Imengine\CropData\Interfaces\NewsPilotCropDataInterface $cropData
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function newsPilotCrop(NewsPilotCropDataInterface $cropData): Imengine
    {
        return static::createMirrorFunction('np_crop', [
            'width'  => $cropData->getWidth(),
            'height' => $cropData->getHeight(),
            'x'      => $cropData->getCropX(),
            'y'      => $cropData->getCropY(),
            'z'      => $cropData->getZoom(),
        ]);
    }

    /**
     * Fetch image with original size
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function original(): Imengine
    {
        return static::createMirrorFunction('original');
    }

    /**
     * Crop image into defined size from center position
     *
     * @param int $width
     * @param int $height
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function thumbnail(int $width, int $height): Imengine
    {
        return static::createMirrorFunction('thumbnail', compact('width', 'height'));
    }

    // Extended mirror functions: functions with extended functionality och the basic mirror functions
    // ======================================================

    /**
     * Crop image into given ratio
     *
     * @param $width
     * @param $ratio
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function ratio(int $width, string $ratio = null): Imengine
    {
        return static::hardCrop($width, ImengineHelper::calculateRatioHeight($width, $ratio ?? '3:2'));
    }

    /**
     * Crop image into given ratio
     *
     * @param \Infomaker\Imengine\CropData\Interfaces\AutoCropDataInterface $cropData
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function autoCrop(AutoCropDataInterface $cropData): Imengine
    {
        [$diffWidth, $diffHeight] = ImengineHelper::calculateCropValues(
            $cropData->getOriginalWidth(),
            $cropData->getOriginalHeight(),
            $cropData->getRatio()
        );

        return static::createMirrorFunction('cropresize', [
            'crop_w' => Format::formatNumberToImengineFloat($diffWidth),
            'crop_h' => Format::formatNumberToImengineFloat($diffHeight),
            'x'      => Format::formatNumberToImengineFloat((1 - $diffWidth) / 2),
            'y'      => Format::formatNumberToImengineFloat((1 - $diffHeight) / 2),
            'width'  => $cropData->getWidth(),
            'height' => ImengineHelper::calculateRatioHeight($cropData->getWidth(), $cropData->getRatio()),
        ]);
    }

    /**
     * Function for fetching the Raw format of the image (May have an unlimited size)
     *
     * @param string $uuid
     * @param int    $quality
     *
     * @return string
     */
    public static function raw(string $uuid, int $quality = null): string
    {
        return (new static(static::getBaseUrl(), [
            'type' => 'primary',
            'q'    => $quality ?? 100,
        ]))->fromUuid($uuid);
    }

    /**
     * Crop image into a square
     *
     * @param int $size
     *
     * @return \Infomaker\Imengine\Imengine
     */
    public static function square(int $size): Imengine
    {
        return static::hardCrop($size, $size);
    }
}
